/* eslint-disable handle-callback-err */
import PouchDB from 'pouchdb-browser'

let databases = {}
let databaseName = ''
let dbKeys = ['db_versions']
let ignoreRoutes = []
let moduleName = ''
let ongoingCachePromises = {}

const getDatabase = (name = null) => {
  if (!databases[name]) {
    databases[name] = new PouchDB(`${databaseName}${name ? `_${name}` : ''}`, {
      adapter: 'idb',
      revs_limit: 1
    })
  }

  return databases[name]
}

/**
 * This function returns a promise(which refreshes cache) if ongoingCachePromise is not set
 * or else(when ongoingCachePromise is set), it returns the same in-progress promise
 */
function getCachePromise (axios, store, cacheKey, cacheRoute, dbVersionsHeaderKey, userId, isPublic = false) {
  let promiseId = `${cacheKey}${userId ? `_${userId}` : ''}`

  if (!ongoingCachePromises[promiseId]) {
    store.commit(`${moduleName}/UPDATE_CACHE_START`)

    let cachePromise = fetchCacheData(axios, store, cacheKey, cacheRoute, dbVersionsHeaderKey, userId, isPublic)

    ongoingCachePromises[promiseId] = cachePromise
      .then(() => {
        resetOngoingCachePromise(promiseId, store)
      })
      .catch((error) => {
        resetOngoingCachePromise(promiseId, store)

        throw error
      })
  }

  return ongoingCachePromises[cacheKey]
}

function resetOngoingCachePromise (cacheKey, store) {
  delete ongoingCachePromises[cacheKey]

  if (Object.keys(ongoingCachePromises).length === 0) {
    store.commit(`${moduleName}/UPDATE_CACHE_END`)
  }
}

function createModule (dbVersions, cacheData, userId, store, axios, dbVersionsHeaderKey, cacheConfigs) {
  let state = {
    db_versions: dbVersions,
    updating_cache: false,
    user_id: userId
  }

  let getters = {
    dbVersions: state => state.db_versions,
    updatingCache: state => state.updating_cache,
    userId: state => state.user_id
  }

  let mutations = {
    'UPDATE_DB_VERSIONS' (state, version) {
      state.db_versions = version
    },
    'UPDATE_CACHE_END' (state) {
      state.updating_cache = false
    },
    'UPDATE_CACHE_START' (state) {
      state.updating_cache = true
    },
    'UPDATE_CACHED_USER_ID' (state, userId) {
      state.user_id = userId
    }
  }

  registerStates(cacheConfigs, state, cacheData)
  registerGetters(cacheConfigs, getters)
  registerMutations(cacheConfigs, mutations)

  return {
    state: state,
    actions: {
      reset: ({getters}) => {
        cacheConfigs.forEach((config) => {
          return getCachePromise(axios, store, config.cacheKey, config.cacheRoute, dbVersionsHeaderKey, getters.userId, config.isCacheRoutePublic)
        })
      },
      updateCachedUserId: ({commit, getters}, id) => {
        if (getters.userId !== id) {
          commit(`UPDATE_CACHED_USER_ID`, id)

          getDatabase(id).allDocs({
            include_docs: true,
            keys: dbKeys
          })
            .then(result => {
              let dbVersions = !result.rows[0].error ? result.rows[0].doc.data : {}
              let cacheData = fetchDbResult(result)

              commit(`UPDATE_DB_VERSIONS`, dbVersions)

              cacheConfigs.forEach(config => {
                let mutation = 'UPDATE_'.concat(config.cacheKey.toUpperCase())

                if (config.isCacheRoutePublic || id) {
                  let version = dbVersions[config.cacheKey]

                  if (typeof version === 'undefined' || !cacheData[config.cacheKey]) {
                    getCachePromise(axios, store, config.cacheKey, config.cacheRoute, dbVersionsHeaderKey, id, config.isCacheRoutePublic)
                  } else {
                    commit(mutation, cacheData[config.cacheKey] || null)
                  }
                } else {
                  dbVersions[config.cacheKey] = null

                  commit(mutation, null)
                }
              })

              commit(`UPDATE_DB_VERSIONS`, dbVersions)
            })
        }
      }
    },
    getters: getters,
    mutations: mutations,
    namespaced: true
  }
}

function registerMutations (cacheConfigs, mutations) {
  cacheConfigs.forEach(config => {
    let mutation = 'UPDATE_'.concat(config.cacheKey.toUpperCase())

    mutations[mutation] = function (state, data) {
      state[config.cacheKey] = data
    }
  })
}

function registerGetters (cacheConfigs, getters) {
  cacheConfigs.forEach(config => {
    getters[config.cacheKey] = state => typeof state[config.cacheKey] === 'undefined' ? [] : state[config.cacheKey]
  })
}

function registerStates (cacheConfigs, state, cacheData) {
  cacheConfigs.forEach(config => {
    state[config.cacheKey] = Object.keys(cacheData).length === 0 ? [] : cacheData[config.cacheKey]
  })
}

function fetchCacheData (axios, store, cacheKey, cacheRoute, dbVersionsHeaderKey, userId, isPublic = false) {
  return new Promise((resolve, reject) => {
    if (isPublic || userId) {
      axios.get(cacheRoute)
        .then(response => {
          let responseData = response.data[cacheKey]
          let dbVersion = getResponseDbVersion(response, dbVersionsHeaderKey, cacheKey)
          let newDbVersions = store.state[`${moduleName}`].db_versions
          let mutation = 'UPDATE_'.concat(cacheKey.toUpperCase())

          newDbVersions[cacheKey] = dbVersion

          store.commit(`${moduleName}/UPDATE_DB_VERSIONS`, newDbVersions)
          store.commit(`${moduleName}/${mutation}`, responseData)

          getDatabase(userId).get(cacheKey)
            .then(doc => {
              doc.data = responseData

              getDatabase(userId).put(doc).catch(error => reject(error))
            })
            .catch(error => {
              getDatabase(userId).put({
                _id: cacheKey,
                data: responseData
              }).catch(error => reject(error))
            })

          getDatabase(userId).get('db_versions')
            .then(doc => {
              doc.data = newDbVersions

              getDatabase(userId).put(doc).catch(error => reject(error))
            })
            .catch(error => {
              getDatabase(userId).put({
                _id: 'db_versions',
                data: newDbVersions
              }).catch(error => reject(error))
            })

          resolve()
        })
        .catch(error => {
          reject(error)
        })
    } else {
      resolve()
    }
  })
}

function fetchDbResult (result) {
  let data = {}

  result.rows.forEach(rs => {
    data[rs.key] = typeof rs.doc === 'undefined' ? {} : rs.doc.data
  })

  return data
}

function getResponseDbVersion (response, dbVersionsHeaderKey, cacheKey) {
  let headerKey = dbVersionsHeaderKey.toLowerCase()

  let dbVersions = typeof response.headers[headerKey] !== 'undefined' ? JSON.parse(response.headers[headerKey]) : {}

  return dbVersions[cacheKey]
}

function initAxiosInterceptor (store, axios, dbVersionsHeaderKey, cacheConfigs) {
  axios.interceptors.response.use((response) => {
    if (!isCacheRoute(response.config.url, response.config.baseURL, cacheConfigs) && !isRouteIgnored(response.config.url, response.config.baseURL)) {
      let userId = store.getters[`${moduleName}/userId`]

      cacheConfigs.forEach(config => {
        if (config.isCacheRoutePublic || userId) {
          let responseVersion = getResponseDbVersion(response, dbVersionsHeaderKey, config.cacheKey)
          let storeVersion = store.getters[`${moduleName}/dbVersions`][config.cacheKey]

          if (typeof storeVersion === 'undefined' || storeVersion !== responseVersion) {
            getCachePromise(axios, store, config.cacheKey, config.cacheRoute, dbVersionsHeaderKey, userId, config.isCacheRoutePublic)
          }
        }
      })
    }

    return response
  })
}

function isRouteIgnored (URL, baseURL) {
  return !!ignoreRoutes.find(route => URL.replace(baseURL, '').match(`^/?${route}$`))
}

function isCacheRoute (cacheURL, baseURL, cacheConfig) {
  return !!cacheConfig.find(config => cacheURL.replace(baseURL, '').match(`^/?${config.cacheRoute}$`))
}

export function initCacheModule (store, axios, userId, dbName, storeModuleName, dbVersionsHeaderKey, cacheConfigs, ignores = []) {
  databaseName = dbName
  ignoreRoutes = ignores
  moduleName = storeModuleName

  let versions = {}

  cacheConfigs.forEach(config => {
    dbKeys.push(config.cacheKey)
    versions[config.cacheKey] = null
  })

  return new Promise((resolve) => {
    getDatabase(userId).allDocs({
      include_docs: true,
      keys: dbKeys
    })
      .then(result => {
        let dbVersions = !result.rows[0].error ? result.rows[0].doc.data : versions
        let cacheData = fetchDbResult(result)

        let module = createModule(dbVersions, cacheData, userId, store, axios, dbVersionsHeaderKey, cacheConfigs)

        store.registerModule(moduleName, module)

        initAxiosInterceptor(store, axios, dbVersionsHeaderKey, cacheConfigs)

        resolve(module)
      })
  })
}
